import astropy.io.fits as fits
import numpy as np
import bpm.image_processing as image_processing
import bpm.image_utils as image_utils

def generate_test_bias_frame(bad_pixel_locations):
    hdr = fits.Header([('BIASSEC', '[95:100, 1:100]'),
                       ('TRIMSEC', '[1:94, 1:100]')])

    overscan_slices = image_utils.get_slices_from_header_section(hdr['BIASSEC'])

    bias_frame = np.round(np.random.normal(1000, 50, (100, 100)))
    bias_frame[bad_pixel_locations] = np.random.normal(5000, 3000)
    bias_frame[overscan_slices] = np.round(np.random.normal(1000, 50))

    return fits.ImageHDU(data=bias_frame, header=hdr)

def generate_test_flat_frame(bad_pixel_locations, image_mean, image_std):
    hdr = fits.Header([('FILTER', 'w'),
                       ('BIASSEC', '[95:100, 1:100]'),
                       ('TRIMSEC', '[1:94, 1:100]')])

    overscan_slices = image_utils.get_slices_from_header_section(hdr['BIASSEC'])

    flat_frame = np.round(np.random.normal(image_mean, image_std, (100,100)))
    flat_frame[bad_pixel_locations] = np.round(np.random.normal(image_mean, 20*image_std))
    flat_frame[overscan_slices] = np.round(np.random.normal(1000, 50))

    return fits.ImageHDU(data=flat_frame, header=hdr)

def generate_test_dark_frame(bad_pixel_locations):
    hdr = fits.Header([('BIASSEC', '[95:100, 1:100]'),
                       ('TRIMSEC', '[1:94, 1:100]'),
                       ('EXPTIME', '10.0')])

    dark_frame = np.round(np.random.normal(30, 5, (100,100)))
    dark_frame[bad_pixel_locations] = 1000

    return fits.ImageHDU(data=dark_frame, header=hdr)

def generate_bad_pixel_locations(x_limit, y_limit, num_pixels):
    bad_pixels_y = np.array([np.random.randint(y_limit) for index in range(0, num_pixels)])
    bad_pixels_x = np.array([np.random.randint(x_limit) for index in range(0, num_pixels)])
    bad_pixels = tuple((bad_pixels_y, bad_pixels_x))

    return bad_pixels



import astropy.io.fits as fits
import os
import datetime
import glob
import logging
import lcogt_logging
import argparse
import bpm.image_processing as image_processing
import numpy as np

logger = logging.getLogger('lco-bpm-maker')

def setup_logging(log_level):
    logger.setLevel(log_level)
    handler = logging.StreamHandler()
    handler.setLevel(log_level)
    handler.setFormatter(lcogt_logging.LCOGTFormatter())
    logger.addHandler(handler)


def parse_args():
    parser = argparse.ArgumentParser(description='Create a bad pixel mask from a set of calibration frames.')
    parser.add_argument('input_directory', help='Input directory of calibration images')
    parser.add_argument('output_directory', help='Output directory for bad pixel mask')
    parser.add_argument('--log-level', dest='log_level', default='INFO', help='Logging level to be displayed',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])
    parser.add_argument('--dark-current-threshold', dest='dark_current_threshold',
                        help='Threshold for pixel dark current when flagging bad pixels in dark frames. Pixels above this will be flagged. Default = 35 [electrons/second]',
                        default=35)
    parser.add_argument('--flat-sigma-threshold', dest='flat_sigma_threshold',
                        help='Number of standard deviations from the median of the combined flat image for a pixel to be flagged as bad. Default = 10',
                        default=10)
    parser.add_argument('--bias-sigma-threshold', dest='bias_sigma_threshold',
                        help='Number of standard deviations from the median of the combined bias image for a pixel to be flagged as bad. Default = 10',
                        default=10)

    args = parser.parse_args()

    return args


def generate_bpm():
    args = parse_args()
    setup_logging(getattr(logging, args.log_level))

    calibration_frames = get_calibration_frames(os.path.normpath(args.input_directory) + '/*.fits')

    if calibration_frames:
        if len(set([frame.header['INSTRUME'] for frame in calibration_frames])) != 1:
            raise RuntimeError("Got calibration frames from more than one camera. Aborting.")

        #check if camera has an overscan region
        camera_has_no_overscan = True
        try:
            bias_section = image_utils.get_slices_from_header_section(calibration_frames[0].header['BIASSEC'])
            camera_has_no_overscan = False
        except:
            logger.warn("Couldn't parse BIASSEC keyword. Using bias frames to determine camera bias level.")

        frames_sorted_by_binning = sort_frames_by_header_values(calibration_frames, 'CCDSUM')

        logger.info("Beginning processing on {num_frames} calibration frames".format(num_frames = len(calibration_frames)))
        for binning in frames_sorted_by_binning.keys():
            bias_level  = image_processing.get_bias_level_from_frames(get_frames_of_type(frames_sorted_by_binning[binning], 'BIAS')) if camera_has_no_overscan else None

            dark_mask = image_processing.process_dark_frames(get_frames_of_type(frames_sorted_by_binning[binning], 'DARK'), int(args.dark_current_threshold), bias_level)
            bias_mask = image_processing.process_bias_frames(get_frames_of_type(frames_sorted_by_binning[binning], 'BIAS'), int(args.bias_sigma_threshold))

            flats_sorted = sort_frames_by_header_values((get_frames_of_type(frames_sorted_by_binning[binning], 'FLAT')), 'FILTER')
            flat_masks = [image_processing.process_flat_frames(flats_sorted[filter], int(args.flat_sigma_threshold), bias_level) for filter in flats_sorted.keys()]

            flat_masks.extend([bias_mask, dark_mask])
            combined_mask = np.sum(np.dstack(flat_masks), axis=2) > 0

            today_date = datetime.datetime.now().strftime("%Y-%m-%d-%H%M%S")
            instrument_code = frames_sorted_by_binning[binning][0].header['INSTRUME']

            header_info = {'OBSTYPE': 'BPM',
                           'DAY-OBS': frames_sorted_by_binning[binning][0].header['DAY-OBS'],
                           'CCDSUM': binning,
                           'SITEID': frames_sorted_by_binning[binning][0].header['SITEID'],
                           'INSTRUME': instrument_code}

            output_filename = os.path.join(args.output_directory, "bpm-{instrument}-{bin_type}-{today}.fits".format(instrument=instrument_code,
                                                                                                                    today=today_date,
                                                                                                                    bin_type=binning.replace(" ", "x")))
            fits.writeto(filename=output_filename,
                         data=combined_mask.astype(np.uint8),
                         header=fits.Header(header_info),
                         checksum=True)

            logger.info("Finished processing. BPM written to {file_path}".format(file_path=output_filename))

    else:
        raise RuntimeError("No calibration frames could be found. Check that the directory contains calibration frames.")


def get_calibration_frames(path_to_frames, calibration_types=['d00', 'f00', 'b00']):
    """
    Given a directory of fits files, return a list of calibration frames
    """
    frames = glob.glob(path_to_frames)
    frames = [fits.open(frame, mode='readonly')[0] for frame in frames if any(obs_type in frame for obs_type in calibration_types)]
    return frames


def get_frames_of_type(frames, observation_type):
    """
    Takes in a list of frames, and returns frames which match the observation
    type provided.
    """
    return [frame for frame in frames if observation_type in frame.header['OBSTYPE']]

def sort_frames_by_header_values(frames, header_keyword):
    """
    Given a set of flat frames and a header keyword, sort the frames by the corresponding
    header values into a form:
    {header_value:[frames_with_header_value]}
    """
    header_values = set([frame.header[header_keyword] for frame in frames])
    return {value: [frame for frame in frames if frame.header[header_keyword] == value]
                    for value in header_values}



import numpy as np
import astropy.stats
import logging
import bpm.image_utils as image_utils

logger = logging.getLogger('lco-bpm-maker')

def process_bias_frames(bias_frames, mask_threshold=10):
    """
    Create bad pixel mask from bias frames.

    :param bias_frames: List of bias frames to be processed
    :param mask_threshold: Number of standard deviations from the median of the combined bias image
    for a pixel to be flagged as bad. (default 10)
    """

    logger.info("Processing {num_frames} bias frames".format(num_frames=len(bias_frames)))
    corrected_frames = []

    for frame in bias_frames:
        image_data = np.float32(frame.data)
        trimsec_section = image_utils.get_slices_from_header_section(frame.header['TRIMSEC'])

        image_data -= np.median(image_data)

        corrected_frames.append(image_data)

    return image_utils.mask_outliers(np.dstack(corrected_frames), mask_threshold)


def process_dark_frames(dark_frames, dark_current_threshold=35, bias_level=None):
    """
    Create bad pixel mask from dark frames.

    :param dark_frames: List of bias frames to be processed
    :param dark_current_threshold: Threshold for pixel dark current when flagging
    bad pixels in dark frames. Pixels above this will be flagged. (default 35 (electrons/second))
    :param bias_level: Bias level for camera (default None)
    If bias_level is None: camera has an overscan region, and bias level will be determined from dark frames
    """

    logger.info("Processing {num_frames} dark frames".format(num_frames=len(dark_frames)))
    corrected_frames = []

    for frame in dark_frames:
        image_data = np.float32(frame.data)
        if bias_level is None:
            overscan_section = image_utils.get_slices_from_header_section(frame.header['BIASSEC'])
            bias_level = np.median(image_data[overscan_section])

        image_data -= bias_level
        image_data /= np.float32(frame.header['EXPTIME'])

        corrected_frames.append(image_data)

    return np.uint8(np.median(np.dstack(corrected_frames), axis=2) > dark_current_threshold)


def process_flat_frames(flat_frames, mask_threshold=10, bias_level=None):
    """
    Create bad pixel mask from flat frames.

    :param flat_frames: List of flat frames to be processed
    :param mask_threshold: Number of standard deviations from the median of the combined flat image for a
    pixel to be flagged as bad. (default 10)
    :param bias_level: Bias level for camera (default None)
    If bias_level is None: camera has an overscan region, and bias level will be determined from flat frames
    """

    logger.info("Processing {num_frames} flat frames taken with filter: {filter}".format(num_frames=len(flat_frames),
                                                                                         filter=flat_frames[0].header['FILTER']))
    corrected_frames = []

    for frame in flat_frames:
        image_data = np.float32(frame.data)
        if bias_level is None:
            overscan_section = image_utils.get_slices_from_header_section(frame.header['BIASSEC'])
            bias_level = np.median(image_data[overscan_section])

        image_data -= bias_level
        image_data /= np.median(image_data)
        
        corrected_frames.append(image_data)

    return image_utils.mask_outliers(np.dstack(corrected_frames), mask_threshold)

def get_bias_level_from_frames(bias_frames):
    """
    Determine camera bias level from a set of bias frames

    :param bias_frames: list of bias frames
    """
    trimsec_section = image_utils.get_slices_from_header_section(bias_frames[0].header['TRIMSEC'])
    bias_level = np.median([np.median(frame.data[trimsec_section]) for frame in bias_frames])

    return bias_level



import numpy as np
import astropy.io.fits as fits
import pytest
import bpm.image_processing as image_processing
import bpm.test.test_utils as test_utils
import bpm.image_utils as image_utils

def test_process_bias_frames():
    bad_pixel_locations = test_utils.generate_bad_pixel_locations(94, 100, 10)
    bias_frames = [test_utils.generate_test_bias_frame(bad_pixel_locations) for index in range(0,10)]

    bias_mask = image_processing.process_bias_frames(bias_frames, mask_threshold=10)
    flagged_pixels = np.where(bias_mask == True)

    assert np.shape(bias_mask) == np.shape(bias_frames[0].data)
    assert set(bad_pixel_locations[0]) == set(flagged_pixels[0])
    assert set(bad_pixel_locations[1]) == set(flagged_pixels[1])


def test_process_dark_frames():
    bad_pixel_locations = test_utils.generate_bad_pixel_locations(94, 100, 10)
    dark_frames = [test_utils.generate_test_dark_frame(bad_pixel_locations) for index in range(0,10)]

    dark_mask = image_processing.process_dark_frames(dark_frames)
    flagged_pixels = np.where(dark_mask==True)

    assert np.shape(dark_mask) == np.shape(dark_frames[0].data)
    assert set(flagged_pixels[0]) == set(bad_pixel_locations[0])
    assert set(flagged_pixels[1]) == set(bad_pixel_locations[1])


def test_process_flat_frames():
    base_image_mean = 22000
    base_image_std = 1000
    bad_pixel_locations = test_utils.generate_bad_pixel_locations(94, 100, 10)

    flat_frames = [test_utils.generate_test_flat_frame(bad_pixel_locations,
                                                       base_image_mean - 2000*index,
                                                       base_image_std - 50*index) for index in range(0,10)]

    flat_mask = image_processing.process_flat_frames(flat_frames, mask_threshold=10)
    flagged_pixels = np.where(flat_mask == True)

    assert np.shape(flat_mask) == np.shape(flat_frames[0].data)
    assert set(bad_pixel_locations[0]) == set(flagged_pixels[0])
    assert set(bad_pixel_locations[1]) == set(flagged_pixels[1])


def test_get_slices_from_header_section():
    test_header_string_1 = '[3100:3135, 1:2048]'
    test_header_string_2 = '[3100:3135,1:2048]'

    assert image_utils.get_slices_from_header_section(test_header_string_1) ==\
           (slice(0, 2048, 1), slice(3099, 3135, 1))

    assert image_utils.get_slices_from_header_section(test_header_string_2) ==\
           (slice(0, 2048, 1), slice(3099, 3135, 1))



import numpy as np
import astropy.stats

def mask_outliers(stacked_frames, mask_threshold=10):
    """
    Mask pixels outside a specified number of standard deviations

    Generate MAD for each pixel - 2D array - pixel_mads
    Generate standard deviation of all pixel MADs - Scalar - std_all_pixels
    Generate median of all pixels MADs - Scalar - median_all_pixels

    Flag any pixels whose std value is outside the median +/- mask_threshold * std_all_pixels

    :param stacked_frames: stack of corrected frames
    :param mask_threshold: standard deviation threshold
    """
    pixel_mads = astropy.stats.median_absolute_deviation(stacked_frames, axis=2)
    std_all_pixels = np.std(pixel_mads)
    median_all_pixels = np.median(pixel_mads)

    outlier_mask = np.logical_or(pixel_mads < median_all_pixels - (mask_threshold * std_all_pixels),
                                 pixel_mads > median_all_pixels + (mask_threshold * std_all_pixels))

    return np.uint8(outlier_mask)


def get_slices_from_header_section(header_section_string):
    """
    Borrowed from BANZAI. Convert FITS header image section value to tuple of slices.

    Example:  '[3100:3135,1:2048]' --> (slice(0, 2048, 1), slice(3099, 3135, 1))
    Note:
    FITS Header image sections are 1-based and indexed by [column, row]
    Numpy arrays are zero-based and indexed by [row, column]

    :param header_string: An image section string in the form "[x1:x2, y1:y2]"
    :return: Row-indexed tuple of slices, (row_slice, col_slice)
    """

    # Strip off the brackets and split the coordinates
    pixel_sections = header_section_string[1:-1].split(',')
    x_slice = split_slice(pixel_sections[0])
    y_slice = split_slice(pixel_sections[1])
    pixel_slices = (y_slice, x_slice)
    return pixel_slices


def split_slice(pixel_section):
    """
    Borrowed from BANZAI. Convert FITS header pixel section to Numpy-friendly
    slice.

    Example: '3100:3135' --> slice(3099, 3135, 1)
    """
    pixels = pixel_section.split(':')
    if int(pixels[1]) > int(pixels[0]):
        pixel_slice = slice(int(pixels[0]) - 1, int(pixels[1]), 1)
    else:
        if int(pixels[1]) == 1:
            pixel_slice = slice(int(pixels[0]) - 1, None, -1)
        else:
            pixel_slice = slice(int(pixels[0]) - 1, int(pixels[1]) - 2, -1)
    return pixel_slice
